// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/* Revision History
 *
 * $Id: PerlInterpreter.cc,v 1.8 2002/09/17 09:38:23 gvacanti Exp $
 *
 */

#include <errstr.h>
#include "perlutils.h"
#include "PerlInterpreter.h"

extern "C" {
#include "EXTERN.h"
#include <perl.h>
}





// ************************************************
// Define static array of lookup table information.
// ************************************************

Perlinterpreter perlInterpreter;


void
Perlinterpreter::initialise( const string & perllib )
{
	initialisePerl( perllib.c_str() );

}


void
Perlinterpreter::addFunctionsToTable( TableT &	table ) const
{
	Lookups lookups[] = {
		{ string( "evalperl" ),		Token( Functor( Functor::C, (void *)evalperl,		int(0)))},
		{ string( "perlcallva" ),	Token( Functor( Functor::C, (void *)perl_call_va,	int(0)))},
		{ string( "perlevalva" ),	Token( Functor( Functor::C, (void *)perl_eval_va, 	int(0)))},
		{ string( "perlgetint" ),	Token( Functor( Functor::C, (void *)perl_get_int, 	int(0)))},
	};

	const unsigned int	numberOfFunctions =

					sizeof lookups /
					sizeof( Lookups );


	// Construct lookup table.
	for( unsigned int i = 0; i < numberOfFunctions; ++i )
	{
		table.Insert( lookups[ i ].name, lookups[ i ].token );

	} // for

	Interpreter::addFunctionsToTable( table );

}


typedef struct {
    char type;       
    void *pdata;
} Out_Param;


Token
Perlinterpreter::EvaluatePerlFunction(
	const Token::BTree &	expression,
	TableT & 		lookupTable ) const
{

	vector<Token::BTree>		listOfParameters;
	Token				token;
	Token				left;
	Token				result;
	int				ival;
	char				sval[ 1024 ];
	double				dval;
	void *				pval;

	char*				str = NULL;
	int				i = 0;
	double				d = 0;
	int  				nret = 0; /* number of return params expected*/
	int 				ii=0;
	void * 				pp;
	Out_Param 			op[20];
	int 				res = 0;

    	djSP;
    	ENTER;
    	SAVETMPS;
    	PUSHMARK(sp);
 
	left = EvaluateExpression(
		expression.left(),
		lookupTable );

	token = expression.right().node();
	listOfParameters = token.asBTreeVector();

	for( unsigned int k = 0; k < listOfParameters.size(); ++k )
	{
		token = EvaluateExpression(
				listOfParameters[ k ],
				lookupTable );

        	switch( token.id() )
        	{
        		case Token::String :
				str = const_cast<char*>(token.asString().c_str());
         			ii = strlen(str);
                		XPUSHs(sv_2mortal(newSVpv(str,ii)));
            			break;
        		case Token::Integer :
               			ii = token.asInt();
               			XPUSHs(sv_2mortal(newSViv(ii)));
            			break;
			case Token::Pointer :
				pp = const_cast<void*>(token.asPointer());
               			XPUSHs(sv_2mortal(newSViv((IV)pp)));
				break;
        		case Token::Real :
				d = token.asReal();
				XPUSHs(sv_2mortal(newSVnv(d)));
            			break;
			default:
				errstr << "Parameter type not supported" << Sas::error( "unsupportedParameterType" );
        	} // switch
	} // while
 
	switch( left.asFunctor().returnType().id() )
	{
		case Token::Integer :
                	op[nret].pdata = &ival;
                	op[nret++].type = 'i';
			break;
		case Token::Real :
                	op[nret].pdata = &dval;
                	op[nret++].type = 'd';
			break;
		case Token::String :
                	op[nret].pdata = &sval;
                	op[nret++].type = 's';
			break;
		case Token::Pointer :
                	op[nret].pdata = &pval;
                	op[nret++].type = 'p';
			break;
		case Token::Undefined :
			break;
		default :
			errstr << "Parameter type not supported" << Sas::error( "unsupportedParameterType" );
			break;
	} // switch

    	PUTBACK;
    	res = perl_call_pv(
		const_cast<char*>( expression.left().node().asString().c_str() ),
		(nret == 0) ? G_DISCARD : (nret == 1) ? G_SCALAR  : G_ARRAY  );
 
 
    	SPAGAIN;
    	if( nret > res )
	{
        	nret = res;
	} // if
 
    	for (i = --nret; i >= 0; i--)
	{
        	switch (op[i].type)
		{
		        case 's':
				str = POPp;
				strcpy((char *)op[i].pdata, str);
				break;
			case 'i':
				ii = POPi;
				*((int *)(op[i].pdata)) = ii;
				break;
			case 'd':
				d = POPn;
				*((double *) (op[i].pdata)) = d;
				break;
			case 'p' :
				// To be Implemented.
				break;
		} // switch
	} // for
   
	FREETMPS ;
	LEAVE ;

	switch( left.asFunctor().returnType().id() )
	{
		case Token::Integer :
			result = ival;
			break;
		case Token::Real :
			result = dval;
			break;
		case Token::String :
			result = sval;
			break;
		case Token::Pointer :
			result = pval;
			break;
		case Token::Undefined :
			break;
		default :
			errstr << "Parameter type not supported" << Sas::error( "unsupportedParameterType" );
	} // switch

	return result;

}


Token
Perlinterpreter::EvaluateFunction(
	const Token::BTree &	expression,
	TableT & 		lookupTable ) const
{

	Token				left;
	Token				result;

	left = EvaluateExpression(
		expression.left(),
		lookupTable );

	switch( left.id() )
	{
		case Token::Defun :
			result = EvaluateDefunFunction( expression, lookupTable );
			break;

		case Token::Function :
			switch( left.asFunctor().type() )
			{
				case Functor::C :
					result = EvaluateCFunction(
							expression,
							lookupTable );
					break;
				case Functor::Perl :
					result = EvaluatePerlFunction(
							expression,
							lookupTable );
					break;
				case Functor::Internal :
					result = EvaluateInternalFunction(
							expression,
							lookupTable );
					break;
				default :
					cerr << "Function type not handled" << endl;
					exit( 1 );
			} // switch
			break;
		case Token::Undefined :
			cerr << "Error: Attempt to invoke a non-existant function: " << endl;
			cerr << expression.left().node().asString().c_str() << endl;
			exit( 1 );
			break;
		default :
			cerr << "Internal error: Unexpected switch/case" << endl;
			exit( 1 );
			break;
	} // switch

	return result;
}

