// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/* Revision History
 *
 * $Id: parsing.cc,v 1.12 2009/07/06 14:20:08 aibarra Exp $
 *
 */


//
// parsing.c
//
//
// HISTORY:
//
//    Date        Author   Modifications
//    --------    ------   -------------
//    16/06/97    MT       Initial Version
//
//
//
//
//
//
//
// DESCRIPTION:
//
//    Implementation file for various parsing functions.
//
//

#include "token.h"

#include <vector>

#include "Vec.h"
#include "parsing.h"
#include "Tracer.h"


static Token::BTree parseRelationalTerm( ScannerT & );
static Token::BTree ParsePlusTerm( ScannerT & );
static Token::BTree ParseMultiplicationTerm( ScannerT & );
static Token::BTree ParseUnaryTerm( ScannerT & );
static Token::BTree ParsePostfixTerm( ScannerT & );
static Token::BTree ParseFunctionTerm( ScannerT & );
static Token::BTree ParseArrayTerm( ScannerT & );
static Token::BTree ParsePrimaryTerm( ScannerT & );
static Token::BTree ParseDefun( ScannerT & );
static Token::BTree ParseLoop( ScannerT & );
static Token::BTree parseCompoundStatement( ScannerT & );
static Token::BTree ParseExitif( ScannerT & );
static Token::BTree ParseReturn( ScannerT & );
static Token::BTree ParseIf( ScannerT & );


Token::BTree
ParseStatement( ScannerT &	scanner )
{

	PROC

	Token		token;
	Token::BTree	tree;

	/* AI gcc4 porting
	 *  deprecated conversion from string constant to ‘char*’ => cast to (char*)
	 */
	if( scanner.NextIs( (char*)"DEFUN" ))
	{
		tree = ParseDefun( scanner );
	}
	else if( scanner.NextIs((char*) "RETURN" ))
	{
		tree = ParseReturn( scanner );
	}
	else if( scanner.NextIs((char*) "IF" ))
	{
		tree = ParseIf( scanner );
	}
	else if( scanner.NextIs((char*)"LOOP" ))
	{
		tree = ParseLoop( scanner );
	}
	else if( scanner.NextIs( (char*)"EXITIF" ))
	{
		tree = ParseExitif( scanner );
	}
	else
	{
		tree = ParseExpression( scanner );
		scanner.Match( ';', token );

	} // if

	RETURN tree;

}


Token::BTree
parseCompoundStatement( ScannerT & scanner )
{
	PROC

	Token			token;
	Token::BTree		statement;
	vector<Token::BTree>	statements;
	Token::BTree		tree;
	Token::BTree		right;

	// ******************************
	// Parse the compound statements.
	// ******************************
	scanner.match( '{' );

	while( !scanner.NextIs( '}' ))
	{
		statement = ParseStatement( scanner );
		statements.push_back( statement );
	} // while

	scanner.match( '}' );

	// **************************************
	// Construct the binary tree of the loop.
	// **************************************
	tree = '{';
	token = statements;
	right.setNode( token );
	tree.setRight( right );

	RETURN tree;
}


Token::BTree
ParseIf( ScannerT & scanner )
{
	PROC

	Token			token;
	Token::BTree		statement;
	Token::BTree		tree;
	vector<Token::BTree>	ifConstruct;

	scanner.Match( (char*)"IF", token );

	// ************************
	// Parse the if expression.
	// ************************
	Token::BTree expression = ParseExpression( scanner );
	ifConstruct.push_back( expression );

	// ************************
	// Parse the If statements.
	// ************************
	scanner.Match( '{', token );

	while( !scanner.NextIs( '}' ))
	{
		statement = ParseStatement( scanner );
		ifConstruct.push_back( statement );
	} // while

	scanner.Match( '}', token );

	// **********************************************
	// Construct the binary tree of the if construct.
	// *********************************************
	tree = Token( ifConstruct, Token::If );

	RETURN tree;
}


Token::BTree
ParseLoop( ScannerT & scanner )
{
	PROC

	Token			token;
	Token::BTree		statement;
	vector<Token::BTree>	statements;
	Token::BTree		tree;

	scanner.Match( (char*)"LOOP", token );

	// *****************************
	// Parse the Loop's statements.
	// *****************************
	scanner.Match( '{', token );

	while( !scanner.NextIs( '}' ))
	{
		statement = ParseStatement( scanner );
		statements.push_back( statement );
	} // while

	scanner.Match( '}', token );

	// **************************************
	// Construct the binary tree of the loop.
	// **************************************
	tree = Token( statements, Token::Loop );

	RETURN tree;
}


Token::BTree
ParseExitif( ScannerT & scanner )
{
	PROC

	Token		token;

	scanner.Match((char*) "EXITIF", token );
	Token::BTree expression = ParseExpression( scanner );
	scanner.Match( ';', token );

	RETURN Token( expression, Token::Exitif );

}



Token::BTree
ParseDefun( ScannerT & scanner )
{
	PROC

	Token			token;
	Token			name;
	Token			argument;
	Vec<Token>		arguments;
	Token::BTree		statement;
	vector<Token::BTree>	statements;
	Token::BTree		tree;
	Token::BTree		defun;
	Token::BTree		dummy;

	// ***********************
	// Parse the DEFUN's name.
	// ***********************
	scanner.Match( (char*)"DEFUN", token );
	scanner.Match( Token::Identifier, name );

	// ****************************
	// Parse the DEFUN's arguments.
	// ****************************
	scanner.Match( '(', token );

	if( !scanner.NextIs( ')' ))
	{
		scanner.Match( Token::Identifier, argument );
		arguments.push_back( argument );

		while( scanner.NextIs( ',' ))
		{
			scanner.Match( ',', token );
			scanner.Match( Token::Identifier, argument );
			arguments.push_back( argument );
		} // while
	} // if

	scanner.Match( ')', token );

	// *****************************
	// Parse the DEFUN's statements.
	// *****************************
	scanner.Match( '{', token );

	while( !scanner.NextIs( '}' ))
	{
		statement = ParseStatement( scanner );
		statements.push_back( statement );
	} // while

	scanner.Match( '}', token );

	// **********************************************
	// Construct the binary tree of the parsed DEFUN.
	// **********************************************
	dummy.setLeft( Token( arguments ));
	dummy.setRight( Token( statements, Token::BinaryTreeVector ));
	defun.setLeft( name );
	defun.setRight( dummy );
	tree = Token( defun, Token::Defun );

	RETURN tree;
}


Token::BTree
ParseReturn( ScannerT & scanner )
{
	PROC

	Token		token;

	scanner.Match((char*) "RETURN", token );
	Token::BTree expression = ParseExpression( scanner );
	scanner.Match( ';', token );

	RETURN Token( expression, Token::Return );

}


Token::BTree
ParseExpression( ScannerT &	scanner )
{

	PROC

	Token		token;
	Token::BTree	left;
	Token::BTree	tree;


	left = ParseAssignmentExpression( scanner );


	while(	scanner.NextIs( ',' ) )
	{
		token = scanner.acceptNext();
		tree.setNode( token );
		tree.setLeft( left );

		tree.setRight( ParseAssignmentExpression( scanner ));

		left = tree;

	} // while

	RETURN left;

}


Token::BTree
ParseAssignmentExpression( ScannerT &	scanner )
{

	PROC

	Token::BTree	tree;
	Token::BTree	left;
	Token		token;


	left = parseRelationalTerm( scanner );

	if( scanner.NextIs( '=' ))
	{
		scanner.Match( '=', token );
		tree.setNode( token );
		tree.setLeft( left );
		tree.setRight( ParseAssignmentExpression( scanner ));
		left = tree;

	} // if


	RETURN left;

}


static Token::BTree
parseRelationalTerm( ScannerT &		scanner )
{
	PROC

	Token		token;
	Token::BTree	left;
	Token::BTree	tree;


	left = ParsePlusTerm( scanner );

	while(	scanner.NextIs( '>' )
		||
		scanner.NextIs( '<' )
		||
		scanner.nextIs( Token::LessEqual )
		||
		scanner.nextIs( Token::GreaterEqual )
		||
		scanner.nextIs( Token::EqualEqual )
		||
		scanner.nextIs( Token::NotEqual ))
	{
		token = scanner.acceptNext();
		tree.setNode( token );
		tree.setLeft( left );

		tree.setRight( ParsePlusTerm( scanner ));

		left = tree;

	} // while

	RETURN left;
}


static Token::BTree
ParsePlusTerm( ScannerT &		scanner )
{

	PROC

	Token		token;
	Token::BTree	left;
	Token::BTree	tree;


	left = ParseMultiplicationTerm( scanner );


	while(	scanner.NextIs( '+' )
		||
		scanner.NextIs( '-' ))
	{
		token = scanner.acceptNext();
		tree.setNode( token );
		tree.setLeft( left );

		tree.setRight( ParseMultiplicationTerm( scanner ));

		left = tree;

	} // while

	RETURN left;

}


static Token::BTree
ParseMultiplicationTerm( ScannerT &		scanner )
{

	PROC

	Token		token;
	Token::BTree	left;
	Token::BTree	tree;


	left = ParseUnaryTerm( scanner );


	while(	scanner.NextIs( '*' )
		||
		scanner.NextIs( '/' ))
	{
		token = scanner.acceptNext();
		tree.setNode( token );
		tree.setLeft( left );

		tree.setRight( ParseUnaryTerm( scanner ));

		left = tree;

	} // while

	RETURN left;

}


static Token::BTree
ParseUnaryTerm( ScannerT &	scanner )
{
	PROC

	Token		token;
	Token::BTree	tree;

	if(
		scanner.NextIs( '&' )
		||
		scanner.NextIs( '!' )

		||
		scanner.NextIs( '+' )

		||
		scanner.NextIs( '-' )

		||
		scanner.NextIs( '*' ) )
	{
		token = scanner.acceptNext();
		tree.setNode( token );

		tree.setLeft( Token( int(0) ) );
		tree.setRight( ParseUnaryTerm( scanner ));
	}
	else
	{
		tree = ParsePostfixTerm( scanner );

	} // if

	RETURN tree;

}


static Token::BTree
ParsePostfixTerm( ScannerT &	scanner )
{
	PROC

	Token::BTree		left;
	Token::BTree		tree;

	left = ParsePrimaryTerm( scanner );

	while( 1 )
	{
		if( scanner.NextIs( '(' ))
		{
			tree = ParseFunctionTerm( scanner );
			tree.setLeft( left );
			left = tree;
		}
		else if( scanner.NextIs( '[' ))
		{
			tree = ParseArrayTerm( scanner );
			tree.setLeft( left );
			left = tree;
		}
		else if(
			scanner.nextIs( Token::PlusPlus )
			||
			scanner.nextIs( Token::MinusMinus ))
		{
			tree = scanner.acceptNext();
			tree.setLeft( left );
			left = tree;
		}
		else
		{
			break;
		} // if
	} // while

	RETURN left;
}


static Token::BTree
ParseFunctionTerm( ScannerT &	scanner )
{
	PROC

	Token::BTree		tree;
	Token::BTree		expression;
	Token::BTree		right;
	Token			token;
	vector<Token::BTree>	listOfExpressions;


	scanner.Match( '(', token );
	tree.setNode( token );

	if( !scanner.NextIs( ')' ))
	{
		expression = ParseAssignmentExpression( scanner );
		listOfExpressions.push_back( expression );

		while( scanner.NextIs( ',' ))
		{
			scanner.Match( ',', token );
			expression = ParseAssignmentExpression( scanner );
			listOfExpressions.push_back( expression );

		} // while
	} // if

	scanner.Match( ')', token );

	token = listOfExpressions;
	right.setNode( token );

	tree.setRight( right );

	RETURN tree;

}


static Token::BTree
ParseArrayTerm( ScannerT &	scanner )
{
	PROC

	Token::BTree		tree;
	Token::BTree		expression;
	Token			token;

	scanner.Match( '[', token );
	tree.setNode( token );
	expression = ParseExpression( scanner );
	tree.setRight( expression );
	scanner.Match( ']', token );

	RETURN tree;

}


static Token::BTree
ParsePrimaryTerm ( ScannerT &	scanner )
{
	PROC

	Token::BTree	tree;
	Token		token;


	switch( scanner.Next().id() )
	{
		case Token::Integer :
		case Token::Real :
		case Token::String :
		case Token::Identifier :
		case Token::QuotedCharacter :

			token = scanner.acceptNext();
			tree.setNode( token );

			break;


		case Token::Character :

			switch( scanner.Next().asCharacter() )
			{
				case '(' :
					scanner.Match( '(', token );
					tree = ParseExpression( scanner );
					scanner.Match( ')', token );

					break;

				case '{' :
					tree = parseCompoundStatement( scanner );

					break;

				default :
					cerr << "Unexpected character: " << scanner.Next().asCharacter() << endl;
					exit( 1 );
			} // switch

			break;


		default :

			cerr << "Error: Invalid Primary Term";

			exit( 1 );
			break;

	} // switch

	RETURN tree;

}


